// CI.java

package com.hp.javaci;

import java.io.*;
import java.util.Collections;
import java.util.LinkedList; /* redo stack */
import java.util.ListIterator;
import java.util.StringTokenizer;
import java.lang.reflect.*;

/** This class provides a replacement for the MPE/iX Command
  * interpreter.  Added features include:
  * <bl>
  *    <li> Extremely fast (sub-second) invocation of Java
  *    <li> Implied run for Java commands
  * </bl>
  *
  * @author Mike Yawn, Erik Vistica
  */

public class CI {
   private static String getVersion() { return "A.02.01"; }
   private static CI ci = new CI();
   public  static CI getCI() { return ci; }

   static { 
      System.loadLibrary("javaci"); 
      new SystemExit();  // hook for System.exit() using JCISecurityManager 
   }

   //MPE subsys = 2
   public static final int CIERROR_975  = ( 975 << 16) + 2;// 63897602;
   public static final int CIERROR_9073 = (9073 << 16) + 2;//594608130;
   public static final int CIERROR_9075 = (9075 << 16) + 2;//594739202;

   private boolean shouldExit = false;

   private String ciPrompt;
   private RedoStack redoStack;
   private int startCPUSecs;
   private int startConnMins;

   private JCIClassLoader loader;
   private String         toolsjar = ":/usr/local/java/latest/lib/tools.jar";

   private BufferedReader stdin;
   private BufferedWriter stdout, stderr;
   public  BufferedReader stdin() { return stdin; } /* used by RedoStack */
  
   private java.util.Properties vmProps;
   private String vmName;

   /* JNI functions that interface to MPE intrinsics */
   private native void   mpeHPCICOMMAND(String cmdimage, int[] parms);
   public native String HPCIGETVAR_STRING(String varname);
   public native int    HPCIGETVAR_INT(String varname);
   public native void   HPERRMSG(int errno);

   /* Constructor */
   public CI() { 
      loader = new JCIClassLoader();
      stdin = new BufferedReader(new InputStreamReader(System.in));
      stdout= new BufferedWriter(new OutputStreamWriter(System.out));
      stderr= new BufferedWriter(new OutputStreamWriter(System.err));
      vmProps = System.getProperties();
      vmName  = vmProps.getProperty("java.vm.name");
      redoStack = new RedoStack();
   }

   /* Main processing loop */
   public void emulateCI() {
      doCICommand("setvar JAVACI_VERSION  '" + getVersion() + "'");

      startCPUSecs     = HPCIGETVAR_INT("HPCPUSECS");
      startConnMins    = HPCIGETVAR_INT("HPCONNMINS");
 
      printBanner();
      printWelcome();

      while(!shouldExit) {
         String input = new String();
         do {
           displayCIPrompt();
           input = input + readLine(stdin);
         } while (input.endsWith("&"));
         input = input.replace('&',' ');

         doCommand(input);
      } 

      //printTrailer();
   }

   /** Called from main loop and from RedoStack
     * to parse and execute a command line.
     */
   public void doCommand(String input) {
      StringTokenizer st = new StringTokenizer(input);
      if (!st.hasMoreTokens()) {       // empty input 
         return;
      }
      
      redoStack.add(input);

      String cmd = st.nextToken();

      // Check for built-in commands 

      // java and friends
      if (cmd.equalsIgnoreCase("java")  ||
          cmd.equalsIgnoreCase("javac") ||
          cmd.equalsIgnoreCase("jar"))  {

         JCommand jcmd = parseJavaArgs(cmd, input.substring(cmd.length()));

         if (!doJavaCommand(jcmd))
            doCICommand(input);
         return;
      }

      if (cmd.equalsIgnoreCase("BYE")) {
         shouldExit = true;
         return;
      }
      if (cmd.equalsIgnoreCase("CHGROUP")) {
         String target = st.nextToken();
         String chgroup = "CHDIR /!HPACCOUNT/" + target;
         doCICommand(chgroup);
         return;
      }
      if (cmd.equalsIgnoreCase("DO")) {
         String token;
         if (st.hasMoreTokens()) {
            token = st.nextToken();
            int offset;
            try {
               offset = Integer.parseInt(token);
               redoStack.doit(offset);
            } catch (NumberFormatException e) {
               redoStack.doit(token);
            }
         } else {
            redoStack.doit(-1);
         }
         return;
      }
      if (cmd.equalsIgnoreCase("EXIT")) {
         shouldExit = true;
         return;
      }
      if (cmd.equalsIgnoreCase("LISTREDO")) {
         redoStack.listredo();
         return;
      }
      if (cmd.equalsIgnoreCase("REDO")) {
         String token;
         if (st.hasMoreTokens()) {
            token = st.nextToken();
            int offset;
            try {
               offset = Integer.parseInt(token);
               redoStack.redo(offset);
            } catch (NumberFormatException e) {
               redoStack.redo(token);
            }
         } else {
            redoStack.redo(-1);
         }
         return;
      }
      if (cmd.equalsIgnoreCase("RESETCLASSLOADER")) {
         //String classpath = HPCIGETVAR_STRING("CLASSPATH");
         //System.setProperty("java.class.path", classpath);
         //System.setProperty("env.class.path", classpath);
         loader = new JCIClassLoader();
         //loader.setClasspath(classpath);
         return;
      }
      if (cmd.equalsIgnoreCase("SETVAR")) {
         int cierror = doCICommand(input);
         if (cierror != 0) return;
         String next = st.nextToken();
         if (next.equalsIgnoreCase("CLASSPATH")) {
            // if we saved the old classpath, we could optimize and
            // not reset the loader if we just appended to the end
            loader = new JCIClassLoader();
         } else if (next.equalsIgnoreCase("HPREDOSIZE")) {
            redoStack.setSize(Integer.parseInt(st.nextToken()));
         }
         return;
      }
/*
      if (cmd.equalsIgnoreCase("jsh")) {
         emulateSh();
         return;
      }
*/

      // Not Java nor a builtin.  Call HPCICOMMAND.
      doCICommand(input);
      return;
   }
   
   public void HPCICOMMAND(String cmd, int[] parms) {
      mpeHPCICOMMAND(cmd+'\r', parms);
   }
   
   private void makeNewJCIClassLoader() {
   
   }

   public int doCICommand(String input) {
      int[] parms = new int[] {
         0, /* cmderror */
         0, /* parmnum  */
         2  /* msglevel */
      };
      HPCICOMMAND(input, parms);

      int cistatus = parms[0];

      if (cistatus == 0) return cistatus;
      // If not found, try Java implied run
      if (cistatus == CIERROR_975) {
         cistatus = 0;
         JCommand jcmd = parseJavaArgs("java", input);
         if (jcmd.needsNewVM()) 
            HPERRMSG(CIERROR_975);
         else if (!doJavaCommand(jcmd))
            HPERRMSG(CIERROR_975);
      } else {
         HPERRMSG(cistatus);
      }
      return cistatus;
   }

   /** Parse a command line that is a Java invocation.  There are 
     * several complexities:
     *  - If cmd = java, arguments to the VM are positional -- anything
     *    prior to the classname is a VM argument, everything after
     *    the classname is an argument to the class.
     *  - for any other command, VM args are prefixed by -J.  If we
     *    were able to pass these through to our own VM, we would
     *    strip the -J, but since any of these will trigger the 
     *    need to create a new VM, we'll leave them as is.
     *  - -hotspot or -classic must be first (not currently enforced)
     *  - Classnames for commndline tools (jar, javac, etc) aren't
     *    specified on the command line, so we have to know them.
     */

   public JCommand parseJavaArgs(String cmd, String raw) {

      JCommand jcmd = new JCommand();
      String token;
      boolean positional = false;
      boolean classSeen  = false;

      String classpath = HPCIGETVAR_STRING("CLASSPATH");
      //avoid NullPointerException when CLASSPATH is not set
      if (classpath == null) classpath = " ";

      boolean needToolsJar = false;

      // Set class-specific fields for command line tools
      if (cmd.compareToIgnoreCase("java") == 0) {
         positional = true;
      } else if (cmd.compareToIgnoreCase("javac") == 0) {
         jcmd.setClassName("com.sun.tools.javac.Main");
         needToolsJar = true;
      } else if (cmd.compareToIgnoreCase("jar") == 0) {
         jcmd.setClassName("com.sun.tools.jar.Main");
         needToolsJar = true;
      }
     
      if (needToolsJar)
         classpath = classpath + toolsjar;

      // TODO: if ! in line, must do substitution
      if (raw.indexOf('!') != -1) {
         raw = evaluateCommand(raw);
      }

      // Parse the arguments
      StringTokenizer st = new StringTokenizer(raw);
      String[] vmArgs;
      String   className;
      String[] classArgs;
      int maxparms = st.countTokens();
      vmArgs = new String[maxparms];
      classArgs = new String[maxparms];
      int numVMArgs = 0, numClassArgs = 0;

      while (st.hasMoreTokens()) {
         token = st.nextToken();
         if (token.startsWith("-")) {
            if (token.equals("-hotspot") && vmName.startsWith("HotSpot"))
               continue;
            if (token.equals("-classic") && vmName.startsWith("Classic"))
               continue;
// ? Should we handle -version here?
            if (token.startsWith("-J") && !positional) {
               vmArgs[numVMArgs++] = token;
               jcmd.setNeedsNewVM();
System.out.println("Will start new VM because of " + token);
               continue;
            }
            if (positional && classSeen)
               classArgs[numClassArgs++] = token;
            else {
               vmArgs[numVMArgs++] = token;
               jcmd.setNeedsNewVM();
System.out.println("Will start new VM because of " + token);
            }
         } else /* no leading - on this argument */ {
            if (positional && !classSeen) {
               classSeen = true;
               jcmd.setClassName(token);
            } else {
               // Must be a class argument
               classArgs[numClassArgs++] = token;
            }
         }
      }
      jcmd.setVMArgs(vmArgs, numVMArgs);
      jcmd.setClassArgs(classArgs, numClassArgs);
      jcmd.setClasspath(classpath);
      return jcmd;
   }
   
   private String evaluateCommand(String raw) {
      int[] parms = new int[] { 0, 0, 2 };
      
      HPCICOMMAND("setvar JAVACI_CMD_RAW  ANYPARM(" + raw + ")", parms);
      return HPCIGETVAR_STRING("JAVACI_CMD_RAW");
   }

   public boolean doJavaCommand(JCommand jcmd) {
   
      boolean succeeded = false;
      String[] args = null; /* need outside try block for javac test */
      if (jcmd.needsNewVM()) return false;
      try {
         String classToRun = jcmd.getClassName();
         if (classToRun == null) return false;

         loader.setClasspath(jcmd.getClasspath());

         Class theClass = Class.forName(classToRun, true, loader);

         // Describe parameters of main() and load the method
         Class[] parameterTypes = new Class[1];
         parameterTypes[0] = String[].class;
// TODO: let JCommand specify methods other than main
         Method main = theClass.getMethod("main", parameterTypes);
         doCICommand("setvar JAVACI_EXIT_STATUS  0");
         args = jcmd.getClassArgs();
         main.invoke(theClass, new Object[] { args } );
         succeeded = true;
      } catch (InvocationTargetException e) {
        if (e.getTargetException() instanceof SystemExitException) {
          //System.out.println("Silently prevent JVM exit");
          succeeded = true;
        } else {
          System.out.println("Should not be here. " + e);
          e.printStackTrace();
          succeeded = false;
        }
      } catch (Exception e) {
         // It probably wasn't a java class after all.
         return false;
      }
      // If we have recompiled a cached class, force reload of
      // all cached classes.  Couldn't do this test following
      // main.invoke() because javac calls System.exit().
      if (jcmd.getClassName().equals("com.sun.tools.javac.Main")) {
         String cls = args[0].substring(0,args[0].indexOf('.'));
         if (loader.isCached(cls)) {
            loader = new JCIClassLoader();
         }
      }
      return succeeded;
   }

   /** Display the CI Prompt string */
   /*package*/ void displayCIPrompt() {
      ciPrompt = HPCIGETVAR_STRING("JAVACI_PROMPT");
      if (ciPrompt == null) {
        ciPrompt = "(JAVACI)" + HPCIGETVAR_STRING("HPPROMPT");
      }
      System.out.print(ciPrompt);
   }

    /** Read a line from $STDIN */
    /*package*/ String readLine(BufferedReader stdin) {
       try {
         String input = stdin.readLine();
         return input;
       } catch (IOException e) {
         // write to stderr may throw IOException, leading to infinity...
         System.out.println("Error reading from $STDIN");
         e.printStackTrace();
         return null;
       }
    }

    /** Print logon banner */
    private void printBanner() {
       String osVersion   = HPCIGETVAR_STRING("HPRELVERSION");
       String userVersion = HPCIGETVAR_STRING("HPVERSION");
       String dateString  = HPCIGETVAR_STRING("HPDATEF");
       String timeString  = HPCIGETVAR_STRING("HPTIMEF");
       System.out.println("HP3000  Release: " + osVersion +
                          "   User Version: "    + userVersion + "   " +
                           dateString + ", " + timeString);
       System.out.println("MPE/iX  JavaCI  " + getVersion() + 
            "  Copyright Hewlett-Packard 2001.  All rights reserved.");
    }

    /** Print logoff banner */
    private void printTrailer() {
       int cpuSeconds     = HPCIGETVAR_INT("HPCPUSECS");
       int connMinutes    = HPCIGETVAR_INT("HPCONNMINS");
       String dateString  = HPCIGETVAR_STRING("HPDATEF");
       String timeString  = HPCIGETVAR_STRING("HPTIMEF");
       System.out.println("CPU=" + (1 + cpuSeconds - startCPUSecs) +
                   ".  Connect=" + (1 + connMinutes - startConnMins) +
                   ". " + dateString + ", " + timeString + ".");
    }

    /** Print welcome message */
    private void printWelcome() {
       // SHOWME > JCIWELCOME
       // skip first 6 lines, display remainder
       // PURGE JCIWELCOME
    }

    public static void main(String[] args) {
      CI.getCI().emulateCI();
    }
}
