/*
 * JCIClassLoader 
 *
 * Provides ability to reset CLASSPATH on the fly, which the
 * standard ClassLoader does not.  This is critical for the
 * JavaCI since user may :SETVAR CLASSPATH at any time, and
 * also may change their cwd (which is important if '.' is
 * part of the CLASSPATH)
 */

package com.hp.javaci;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry; 
import java.util.zip.ZipInputStream;

public class JCIClassLoader extends java.lang.ClassLoader {

   public static int MAXCLASSSIZE = 256 * 1024;

   class LoadedClass {
      long   lastModified;
      String fullPath;
      Class  klass;

      public LoadedClass(Class klass, String path, long time) {
         this.klass   = klass;
         fullPath     = path;
         lastModified = time;
      }
   } /* inner class LoadedClass */

   private HashMap   classes = new HashMap();
   private ArrayList paths;
   private boolean   resolveIt;

   /** Used to decide whether to nuke the classloaded and force reload */
   public boolean isCached(String classname) {
      return classes.containsKey(classname);
   }

   public Class findClass(String classname) throws ClassNotFoundException {
      return loadClass(classname, true);
   }

   public synchronized Class loadClass(String className, boolean resolveIt)
                     throws ClassNotFoundException {

      this.resolveIt = resolveIt;

// ***********************************************************
// This entire block of code is never used, because the 'real'
// classloader is caching classes and we never are invoked on
// classes that have previously been loaded.  Keeping the code
// around in case we find a different way to get to this point.
/*
      if (classes.containsKey(className)) {
         // We have previously loaded this class.  Has it changed?
         LoadedClass oldClass = (LoadedClass)classes.get(className);
         File newData = new File(oldClass.fullPath);
         if (newData.lastModified() == oldClass.lastModified) {
            System.out.println("Class cached, hasn't changed.");
            return oldClass.klass;
         } else {
            System.out.println("Class cached, but has been modified");
            if (oldClass.fullPath.endsWith(".jar"))
               return loadClassFromJar(oldClass.fullPath, className);
            else
               return loadClassFromFile(oldClass.fullPath, className);
         }
      }
*/
// **************************************************************

      // Class was not previously loaded by this classloader.
      // Try loading it now.
      Class result;
      byte[] classData;
      Iterator pi = paths.iterator();

      while (pi.hasNext()) {
         String path = (String) pi.next();
         //System.out.println("emulateCI.while.path.class=" + path +
         // " " + className);
         if (path.endsWith(".jar")) 
            result = loadClassFromJar(path, className);
         else
            result = loadClassFromFile(path, className);
 
         if (result == null) {
            continue;
         } else {
            //System.out.println("emulateCI.while.return");
            return result;
         }
      }
      //System.out.println("emulateCI.while.done");

      // We couldn't find it.  Try the system classloader
      result = super.findSystemClass(className);
      // if not found, exception thrown and we don't reach here
      //System.out.println(className + " loaded by super.findSystemClass()");
      return result;
   }

   public Class loadClassFromFile(String path, String className) {
      byte[] classData = null;
      File f = null;
      try {
         f = new File(path + "/" + className + ".class");
         if (f.exists()) {
            FileInputStream fis = new FileInputStream(f);
            classData = new byte[(int)f.length()];
            fis.read(classData);
         } else {
            return null;
         }
      } catch (Exception e2) {
        e2.printStackTrace();
        //if a finally block is added later, control will transfer there
        return null;
      }

      // Should have class bytes loaded by this point
      Class result = defineClass(null, classData, 0, classData.length);
      if (resolveIt) resolveClass(result);
      LoadedClass newClass = new LoadedClass(result, 
                             f.getAbsolutePath(), f.lastModified());
      classes.put(className, newClass);
      return result;
   }

   public Class loadClassFromJar(String jarfile, String className) {
      byte[] classData = null;
      File f = null;
      JarFile jf = null;
      //long start = 0, end = 0;
      //start = System.currentTimeMillis();
      try {
         int len = 0;
         f = new File(jarfile);
         if (f.exists()) {
            jf = new JarFile(f);
            className = className.replace('.','/');
            ZipEntry ze = jf.getEntry(className + ".class");
            if (ze == null) return null; 
            InputStream zis = jf.getInputStream(ze); 
            int size = (int) ze.getSize();
            if (size == -1) size = MAXCLASSSIZE;
            classData = new byte[size];
            int chunkLen = 0;
            while ((chunkLen = zis.read(classData, len, size)) != -1) {
               len += chunkLen;
               size -= chunkLen;
               if (size == 0) break;
            }
            Class result = defineClass(className, classData, 0, len);
            if (resolveIt) resolveClass(result);
            LoadedClass newClass = new LoadedClass(result, 
                    f.getAbsolutePath(), f.lastModified());
            classes.put(className, newClass);
            //end = System.currentTimeMillis();
            //System.out.println((end - start) + " ms to load " + 
            //                    className + " from jar");
             return result;
         } else { 
             /* jarfile does not exist */ 
             return null;
         }
      } catch (IOException e) { 
         return null; 
      }
   }

   public void setClasspath(String cp) {
      paths = new ArrayList();
      /* Portable thing to do would be to get the file.separator
       * property.  But why should this thing be portable?
       */
      StringTokenizer st = new StringTokenizer(cp, ":");
      while (st.hasMoreTokens()) {
         paths.add(st.nextToken());
      }
   }
}
