// RedoStack.java

package com.hp.javaci;

import java.util.ArrayList;
import java.util.ListIterator;

/** This class emulates the MPE/iX Command Interpreter redo
  * stack functionality
  */
public class RedoStack
{
   private int       redoStackSize = 50;
   private int       firstSeqNum = 1;
   private ArrayList redoStack = new ArrayList(redoStackSize);
   private static CI ci;

// Tried this as static initializer, and also as constructor --
// always returns null!
   static {
      ci = CI.getCI();
   }

   /** Add a command string into the redo stack.  If the 
     * stack exceeds the maximum size, the oldest entries are
     * deleted.
     */
   public void add(String cmd) {
      redoStack.add(cmd);
      while (redoStack.size() > redoStackSize) {
         redoStack.remove(0);
         firstSeqNum++;
      }
   }
   
   /** Redo the most recent command in the stack beginning with
     * the pattern
     */
   public void doit(String pattern) {
      pop(); /* remove 'DO' from the stack */
      ListIterator stack = redoStack.listIterator(redoStack.size());
      if (ci == null) ci = CI.getCI();
      while (stack.hasPrevious()) {
         String rcmd = (String)stack.previous();
         if (rcmd.toUpperCase().startsWith(pattern.toUpperCase())) {
            ci.displayCIPrompt();
            System.out.println(rcmd);
            ci.doCommand(rcmd);
            return;
         }
      }
      ci.HPERRMSG(ci.CIERROR_9075); 
   }

   /** Redo the command at the indicated position.  Positive numbers
     * are sequential command numbers since JavaCI was invoked; 
     * negative numbers are offsets from the current command.
     */
   public void doit(int position) {
      pop(); /* remove 'DO' from the stack */
      if (ci == null) ci = CI.getCI();
      String rcmd;
      if (position > 0) {
         if ((position < firstSeqNum) ||
             (position > firstSeqNum + redoStack.size()) ) {
            ci.HPERRMSG(ci.CIERROR_9073);
            return; 
         }
         rcmd = (String)redoStack.get(position - firstSeqNum);
      } else {
         if (Math.abs(position) > redoStack.size()) {
            ci.HPERRMSG(ci.CIERROR_9073);
            return; 
         }
         rcmd = (String)redoStack.get(redoStack.size() + position);
      }
      ci.displayCIPrompt();
      System.out.println(rcmd);
      ci.doCommand(rcmd);
   }

   /** Apply edits to the command */
   public String editloop(String command) {
      String edit = null;
      String invokeEdit = null;
      while (true) {
         System.out.println(command);
         edit = ci.readLine(ci.stdin());
         if (edit.length() == 0) {
            return command;
         }
         
         command = evaluateCommandEdit(command, edit);
      }
   }
   
   private String evaluateCommandEdit(String raw, String edit) {
      int[] parms = new int[] { 0, 0, 2 };
      
      ci.HPCICOMMAND("setvar JAVACI_REDO_RAW  ANYPARM(" + raw  + ")", parms);
      ci.HPCICOMMAND("setvar JAVACI_REDO_EDIT ANYPARM(" + edit + ")", parms);
      ci.HPCICOMMAND("setvar JAVACI_REDO_CMD  EDIT(JAVACI_REDO_RAW, JAVACI_REDO_EDIT)", parms);
        
      return ci.HPCIGETVAR_STRING("JAVACI_REDO_CMD");
   }

   /** Display the contents of the redo stack */
   public void listredo() {
      ListIterator cmdstack = redoStack.listIterator();
      int depth = firstSeqNum;
      while (cmdstack.hasNext()) {
//TODO: align )'s
         System.out.println("   " + depth++ + ") " + cmdstack.next());
      }
   }

   /** Remove the most recent entry from the redo stack.  Used by
     * DO and REDO, since they aren't kept in the stack.
     */
   public void pop() {
      redoStack.remove(redoStack.size() - 1);
   }

   /** Redo a command (pattern match version) */
   public void redo(String pattern) {
      pop(); /* remove 'REDO' from the stack */
      ListIterator stack = redoStack.listIterator(redoStack.size());
      if (ci == null) ci = CI.getCI();
      while (stack.hasPrevious()) {
         String rcmd = (String)stack.previous();
         if (rcmd.toUpperCase().startsWith(pattern.toUpperCase())) {
            rcmd = editloop(rcmd);
            ci.displayCIPrompt();
            System.out.println(rcmd);
            ci.doCommand(rcmd);
            return;
         }
      }
      ci.HPERRMSG(ci.CIERROR_9075);
   }

   /** Redo a command (positional version */
   public void redo(int position) {
      pop(); /* remove 'REDO' from the stack */
      if (ci == null) ci = CI.getCI();
      String rcmd;
      if (position > 0) {
         if ((position < firstSeqNum) ||
             (position > firstSeqNum + redoStack.size()) ) {
            ci.HPERRMSG(ci.CIERROR_9073);
            return; 
         }
         rcmd = ((String)redoStack.get(position - firstSeqNum));
      } else {
         if (Math.abs(position) > redoStack.size()) {
            ci.HPERRMSG(ci.CIERROR_9073);
            return; 
         }
         rcmd = (String)redoStack.get(redoStack.size() + position);
      }
      rcmd = editloop(rcmd);
      ci.displayCIPrompt();
      System.out.println(rcmd);
      ci.doCommand(rcmd);
   }

   /** Set the size of the redo stack */
   public void setSize(int size) {
      redoStackSize = size;
      while (redoStack.size() > redoStackSize) {
         redoStack.remove(0);
         firstSeqNum++;
      }
   }
}
